# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/09/2017"

import unittest
from ..sinogramselection import getSelection, getAngles, selectionIsValid
import numpy

import logging
_logger = logging.getLogger("unitests")


class testSinogramSelection(unittest.TestCase):
    """
    Test some function of test_utils
    """
    def testSelectionIsValid(self):
        """Make sure the selectionIsValid function is working"""
        self.assertTrue(selectionIsValid("3"))
        self.assertTrue(selectionIsValid("3; 7"))
        self.assertTrue(selectionIsValid("23:56"))
        self.assertTrue(selectionIsValid(":3"))
        self.assertTrue(selectionIsValid("2:3; 5:89"))
        self.assertTrue(selectionIsValid("1:3;:6"))
        self.assertTrue(selectionIsValid("1:3;"))

    def testSelection(self):
        p = numpy.arange(10)
        self.assertTrue(
            numpy.array_equal(getSelection(projections=p, selection='2; 4; 5'),
                              numpy.append(numpy.append(p[2], p[4]), p[5]))
        )

        self.assertTrue(
            numpy.array_equal(getSelection(projections=p, selection='1:3'),
                              p[1:3])
        )
        self.assertTrue(
            numpy.array_equal(getSelection(projections=p, selection='3:6; 7:9'),
                              numpy.append(p[3:6], p[7:9]))
        )

        with self.assertRaises(Exception):
            getSelection(projections=None, selection='0')

        with self.assertRaises(Exception):
            getSelection(projections=p, selection='20')

    def testGetAngles(self):
        minAngle = 0
        maxAngle = numpy.pi
        nbAngles = 20

        self.assertTrue(
            numpy.array_equal(getAngles(minAngle=minAngle,
                                        maxAngle=maxAngle,
                                        nbAngles=nbAngles,
                                        selection='',
                                        lastAngleEqualFirst=True,
                                        acquiInverted=False),
                              numpy.linspace(minAngle, maxAngle, nbAngles)
                              )
        )

        self.assertTrue(
            numpy.array_equal(getAngles(minAngle=minAngle,
                                        maxAngle=maxAngle,
                                        nbAngles=nbAngles,
                                        selection='',
                                        lastAngleEqualFirst=False,
                                        acquiInverted=False),
                              numpy.linspace(minAngle, maxAngle, nbAngles+1)[:-1]
                              )
        )

        self.assertTrue(
            numpy.array_equal(getAngles(minAngle=minAngle,
                                        maxAngle=maxAngle,
                                        nbAngles=nbAngles,
                                        selection='0:10',
                                        lastAngleEqualFirst=True,
                                        acquiInverted=False),
                              numpy.linspace(minAngle, maxAngle, nbAngles)[0:10]
                              )
        )


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(testSinogramSelection))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""group some functions used for reconstruction """


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"


import numpy as np
import os
import freeart
from freeart.unitsystem import metricsystem
from freeart.utils import outgoingrayalgorithm
from fabio.edfimage import edfimage
from fisx import Material
import logging
_logger = logging.getLogger("reconstrutils")


def decreaseMatSize(mat):
    """
    will decrease the matrice size if possible where one dimension has only
    one element
    """
    if(len(mat.shape) == 3):
        if(mat.shape[0] == 1):
            mat.shape = (mat.shape[1], mat.shape[2])
        elif (mat.shape[2] == 1):
            mat.shape = (mat.shape[0], mat.shape[1])

    if(len(mat.shape) == 4):
        if(mat.shape[0] == 1):
            mat.shape = (mat.shape[1], mat.shape[2], mat.shape[3])
        if(mat.shape[3] == 1):
            mat.shape = (mat.shape[0], mat.shape[1], mat.shape[2])

    return mat


def increaseMatSize(mat, firstPos):
    """
    Simple function which will add a dimension to the current matrice

    :param mat: the matrice for which we want to remove a dimension
    :param firstPos: if true remove the first position else remove the last
    """
    if(len(mat.shape) == 2):
        if(firstPos):
            mat.shape = (1, mat.shape[0], mat.shape[1])
        else:
            mat.shape = (mat.shape[0], mat.shape[1], 1)

    return mat


def saveMatrix(data, fileName, overwrite=False):
    """
    will create an edf file to store the given data

    :param data: the data to save
    :param fileName: the localisation to know where to save the data
    """
    savePhantom(data, fileName, overwrite)


def saveSinogram(data, fileName, overwrite=False):
    """
    will create an edf file to store the given data

    :param data: the data to save
    :param fileName: the localisation to know where to save the data
    """
    savePhantom(data, fileName, overwrite)


def savePhantom(data, fileName, overwrite=False):
    """
    will create an edf file to store the given data

    :param data: the data to save
    :param fileName: the localisation to know where to save the data
    """
    if(os.path.isfile(fileName)):
        if overwrite is True:
            _logger.info("overxtriting the file : " + str(fileName))
        else:
            _logger.info("file " + str(fileName) + "already exists")
            return

    phantomToSave = data.copy()
    phantomToSave = decreaseMatSize(phantomToSave)
    edf_writer = edfimage(data=phantomToSave)
    edf_writer.write(fileName)
    edf_writer = None


def LoadEdf_2D(fName, frame=0):
    """
    Load data from the given edf file

    :param fName: the name of the file
    :param frame: if multiple frame file, will use theis parameter to load a
                  specific frame
    """
    edfreader = edfimage()
    edfreader.read(fName)
    if edfreader.nframes == 1:
        return edfreader.getData()
    else:
        return edfreader.getframe(frame).data


def readSinoFile(fname):
    """
    Read the sinogram contained in a simple file (txt)

    :return:
        * sinoData : numpy array of float 64 loaded fron the file
        * sinoAngles: list of angles of acquisition
    """
    ind = 0
    fi = open(fname)
    slicesNb = int(fi.readline())
    anglesNb = int(fi.readline())
    pixelsNb = int(fi.readline())

    angles = fi.readline().split()
    sinoAngles = np.array(angles, dtype=np.float64)

    sinoData = np.zeros((slicesNb, anglesNb, pixelsNb), np.float64)

    for line in fi:
        oneLineDat = line.split()
        sinoData[0, ind, :] = oneLineDat
        ind = ind + 1

    _logger.info("sliceNb =" + slicesNb)
    _logger.info("anglesNb =" + anglesNb)
    _logger.info("pixelsNb =" + pixelsNb)

    return (sinoData, sinoAngles)


def savePhantomToTxt(fname, x, fmt="%1.6e", delimiter=' '):
    saveMatrixToTxt(fname, x, fmt, delimiter)


def saveSinogramToTxt(fname, x, fmt="%1.6e", delimiter=' '):
    assert(len(x.shape) in [2, 3])

    with open(fname, 'w') as fh:
        assert(len(x.shape) == 3)
        assert(x.shape[0] == 1)
        fh.write("%d\n" % x.shape[0])

        fh.write("%d\n" % x.shape[1])
        fh.write("%d\n" % x.shape[2])

        nbrow = len(x)
        currentLine = 0
        for row in x[0]:
            line = ' '
            line = line + delimiter.join(fmt % value for value in row)
            if(currentLine != (nbrow-1)):
                fh.write(line + '\n')
            else:
                # write the last line of the file
                fh.write(line)

            currentLine += 1


def saveMatrixToTxt(fname, x, fmt="%1.6e", delimiter=' '):
    """
    will store the data to a file
    """
    assert(len(x.shape) in [2, 3])

    with open(fname, 'w') as fh:
        fh.write("%d\n" % x.shape[0])
        fh.write("%d\n" % x.shape[1])
        if(len(x.shape) > 2):
            assert(x.shape[2] == 1)
            fh.write("%d\n" % x.shape[2])

        nbrow = len(x)
        currentLine = 0
        for row in x:
            line = ' '
            line = line + delimiter.join(fmt % value for value in row)
            if(currentLine != (nbrow-1)):
                fh.write(line + '\n')
            else:
                # write the last line of the file
                fh.write(line)

            currentLine += 1


def readMatrix2D(fname, fmt="%1.6e", delimiter=' '):
    """
    read a 2D matrix stored on a txt file
    """
    with open(fname) as fi:
        # get the matrice dimension from the header
        x_shape = int(fi.readline())
        y_shape = int(fi.readline())
        z_shape = int(fi.readline())
        assert(z_shape == 1)

        matrice = np.ndarray(shape=(x_shape, y_shape, z_shape))

        # read the matrice and affecte the readed value
        x = 0
        for line in fi:
            values = line.split()
            y = 0
            for val in values:
                matrice[x][y][0] = np.float64(val)
                y = y+1

            x = x+1

        return matrice


def makeFreeARTFluoSinogram(phantom, absMat, selfAbsMat, numAngle, detSetup,
                            oversampling, beamCalcMeth, outRayPtCalcMeth,
                            minAngle=0.0, maxAngle=2.0 * np.pi,
                            subdivisionValue=3, turnOffSolidAngle=False,
                            I0=1.0, voxelSize=1.0*metricsystem.centimeter):
    """
    Generate the fluorescence sinogram

    :param phantom: The initial phantom (density * probability of emission of photon...)
    :param absMat: absorption of the incoming. Aka :math:`{\mu}_{E_0}{\\rho}x`

    :return:
        * sinogram (numpy array of float64 )
        * angles list of angles of projection

    .. warning:: all angle values must be given in rad
    """
    assert(len(phantom.shape) == len(absMat.shape))
    assert(len(phantom.shape) == len(selfAbsMat.shape))
    assert(len(phantom.shape) == 3)
    assert(phantom.shape[2] == 1)
    assert(absMat.shape[2] == 1)
    assert(selfAbsMat.shape[2] == 1)
    # create the sinogram
    al = freeart.FluoFwdProjection(phMatr=phantom,
                                   expSetUp=detSetup,
                                   absorpMatr=absMat,
                                   selfAbsorpMatrix=selfAbsMat,
                                   angleList=None,
                                   minAngle=minAngle,
                                   maxAngle=maxAngle,
                                   anglesNb=numAngle)

    al.setOverSampling(oversampling)
    al.setRayPointCalculationMethod(beamCalcMeth)
    al.setOutgoingRayAlgorithm(outRayPtCalcMeth)
    al.turnOffSolidAngle(turnOffSolidAngle)
    al.setVoxelSize(voxelSize)

    if outRayPtCalcMeth == outgoingrayalgorithm.matriceSubdivision:
        al.setSubdivisionSelfAbsMat(subdivisionValue)
    al.setI0(I0)
    # Make sure we can assure repeatability. This will set the seed to 0
    al.setRandSeedToZero(True)

    sinogram = None
    sinogram, angles = al.makeSinogram()

    return sinogram, angles


def convertToFisxMaterial(materialName, material):
    """
    convert a material defined in a dictionnary (from tomoGUI for example.
    Used to be saved and loaded ) in a fisx Material. Won't do the registration

    :param materialName: the name to attribute to the fisx material
    :param material: the material defined in a dictionary

    :return: the fisx material

    initial material structure looks like :
    materialName = "My_mat"
    material = {'Comment':"No comment",
             'CompoundList':['Cr', 'Fe', 'Ni'],
             'CompoundFraction':[18.37, 69.28, 12.35],
             'Density':1.0,
             'Thickness':1.0}

    Final structure is a fisx Material :
    steel = {"Cr":18.37,
         "Fe":69.28, # calculated by subtracting the sum of all other elements
         "Ni":12.35 }
    material = Material("My_mat", 1.0, 1.0)
    material.setComposition(steel)
    """
    assert('CompoundList' in material)
    assert('CompoundFraction' in material)
    assert('Density' in material)
    assert('Thickness' in material)
    # if we have only one CompoundList
    if type(material['CompoundList']) is str:
        composition = {}
        composition[material['CompoundList']] = material['CompoundFraction']

    else:
        assert(len(material['CompoundList']) == len(material['CompoundFraction']))
        composition = {}
        for iCompound in np.arange(0, len(material['CompoundList'])):
            composition[material['CompoundList'][iCompound]] = material['CompoundFraction'][iCompound]
        if len(composition) is 0:
            raise ValueError('Material with empry composition found. '
                             'Can\'t convert material to fisx')

    material = Material(materialName,
                        material['Density'],
                        material['Thickness'])
    material.setComposition(composition)

    return material


def getFisxMatName(materialName):
    """Small hack to make sure the given material name is not existing yet on
    fisx"""
    return '_' + materialName


def tryToFindPattern(filePath, groupType=None):
    """Basically try to fit with pymca output files

    :param filePath: the file we want to find pattern for.

    :return: the files we can group and the type of group we can process
    """

    # try to fit with pymca files
    if not groupType or groupType == 'pymca':
        pattern = 'ANALYSIS/det'
        if pattern in filePath:
            pymcaFiles = []
            if filePath.count(pattern) > 1:
                mess = """found more than one pattern in the folder path.
                    Can t determine any"""
                _logger.info(mess)
                return

            assert(len(pattern) < len(filePath)+2)
            indexPattern = filePath.find(pattern)
            # get detxx value
            detPatternStart = indexPattern+len(pattern)-3
            detPatternEnd = indexPattern+len(pattern)+2
            currentDet = filePath[detPatternStart: detPatternEnd]
            # try replacing det xx by all possible values
            for iDet in range(100):
                if iDet < 10:
                    detID = 'det0' + str(iDet)
                else:
                    detID = 'det' + str(iDet)

                potentialFileID = filePath.replace(currentDet, detID)
                if os.path.isfile(potentialFileID):
                    pymcaFiles.append(potentialFileID)

            if len(pymcaFiles) > 0:
                return pymcaFiles, 'pymca'

    return [], None



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import unittest
import numpy as np

import freeart
from freeart.utils import unittestpercentage, reconstrutils, testutils, raypointsmethod, outgoingrayalgorithm
from freeart.unitsystem import metricsystem

import logging
_logger = logging.getLogger("unitests")

class testRayParameters(unittestpercentage.testPercentage):
    """The class to test some of the parameters relative to rays"""

    # detector properties
    detectorRadius  = 10.0 *metricsystem.cm
    detectorDistance = 1000 *metricsystem.cm
    # Dimensions of the phantom to test
    phDimX = 9 # voxels
    phDimY = 9 # voxels
    phDimZ = 1 # voxels

    xpos   = 4 # voxels
    ypos   = 4 # voxels
    zpos   = 0 # voxels

    origin = (0.0, 0.0, 0.0)

    I0 = 17.5e4 

    voxelPosInPhantomRef = (xpos, ypos, zpos)
    voxelPosInExpRef = testutils.toExperimentationRef( voxelPosInPhantomRef, phDimX, phDimY, phDimZ)

    # we are setting an hight value to get close from the theoretical values
    oversampling = 10
    # For this test case we are skipping the attenuation of the incoming beam and only focusing on the outgoing
    interactionProba = 1.0

    def testRayIntensityProjection(self):           
        """
        Check if the intensity of the source is well taking into account during the projection
        """

        _logger.info("running testRayIntensity")

        # The outgoing beam attenuation to consider for this test case
        incomingAbsorptionFactor = 0.5 
        outgoingAbsorptionFactor = 0.25

        methodGBeams = raypointsmethod.withInterpolation
        optionsOutgoing = [ outgoingrayalgorithm.createOneRayPerSamplePoint, 
            outgoingrayalgorithm.matriceSubdivision, 
            outgoingrayalgorithm.rawApproximation ] 

        detAngle = 0.0
        # we are in the trigonometric reference
        detPos = (np.sin(detAngle)*self.detectorDistance, np.cos(detAngle)*self.detectorDistance, self.zpos)
        detSetup = [(detPos, self.detectorRadius)]

        ### RUN the FREE ART EXPERIMENTATION
        # create the simple phantom
        phantom    = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        absMat     = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        selfAbsMat = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)

        # dans ce cas dans le repere de freeART les lignes se trouve en (:, -5:-7)
        phantom[self.ypos, self.xpos, self.zpos]    = self.interactionProba
        absMat[self.ypos, self.xpos, self.zpos]     = incomingAbsorptionFactor
        selfAbsMat[self.ypos, self.xpos, self.zpos] = outgoingAbsorptionFactor

        for optionOutgoing in optionsOutgoing :

            sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
                phantom, absMat, selfAbsMat, 1, detSetup, self.oversampling, methodGBeams, optionOutgoing, I0=self.I0 )

            # The momentum when we are going trought the n voxel ( n == width )
            I_FreeARTValue_NVoxels = sinogram[0][0][self.xpos]

            ### GET THE THEORETICAL VALUES
            I_theoreticalValue_NVoxels = testutils.getFluoTheoreticalValue(
                detPos,
                self.detectorRadius,
                self.voxelPosInExpRef,           # < voxel position in the freeART reference
                incomingAbsorptionFactor,        # < mass attenuation for the incoming beam
                outgoingAbsorptionFactor,        # < mass attenuation for the outgoing beam
                np.pi/2.0,                       # < angle of entry of the incoming beam
                np.pi/2.0 - detAngle,            # < outgoing beam angle to the sample
                1.0,                             # < tau
                1.0,                             # < material width (in cm)
                I0=self.I0
                )


            _logger.info("theoreticalValue n voxel 0 s= " + str(I_theoreticalValue_NVoxels))
            _logger.info("FreeART value 0 = " + str(I_FreeARTValue_NVoxels))
            _logger.info("error = " + str(self.getError(I_theoreticalValue_NVoxels, I_FreeARTValue_NVoxels)) + " %")
            
            self.assertErrorPercentageAlmostEqual( I_theoreticalValue_NVoxels, I_FreeARTValue_NVoxels, 5 )

    def testRayIntensityReconstruction(self):
        """Check if the intensity of the source is well taking into account during the projection and the reconstruction"""

        _logger.info("running testRayIntensity")

        # The outgoing beam attenuation to consider for this test case
        incomingAbsorptionFactor = 0.5 
        outgoingAbsorptionFactor = 0.25

        methodGBeams = raypointsmethod.withInterpolation
        optionsOutgoing = [ outgoingrayalgorithm.createOneRayPerSamplePoint,
            outgoingrayalgorithm.matriceSubdivision, 
            outgoingrayalgorithm.rawApproximation ] 

        detAngle = 0.0
        # we are in the trigonometric reference
        detPos = (np.sin(detAngle)*self.detectorDistance, np.cos(detAngle)*self.detectorDistance, self.zpos)
        detSetup = [(detPos, self.detectorRadius)]

        ### RUN the FREE ART EXPERIMENTATION
        # create the simple phantom
        initial_phantom    = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        absMat     = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        selfAbsMat = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)

        # dans ce cas dans le repere de freeART les lignes se trouve en (:, -5:-7)
        initial_phantom[self.ypos, self.xpos, self.zpos]    = self.interactionProba
        absMat[self.ypos, self.xpos, self.zpos]     = incomingAbsorptionFactor
        selfAbsMat[self.ypos, self.xpos, self.zpos] = outgoingAbsorptionFactor

        for optionOutgoing in optionsOutgoing :

            sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
                initial_phantom, absMat, selfAbsMat, 60, detSetup, self.oversampling, methodGBeams, optionOutgoing, I0=self.I0 )


           # make the reconstruction with the absMat
            alRecons = freeart.FluoBckProjection(sinoDat=sinogram, sinoAngles=angles, expSetUp=detSetup, absorp=absMat, selfAbsorp=selfAbsMat)

            alRecons.setOverSampling(self.oversampling)
            alRecons.setDampingFactor(0.1)
            alRecons.setRayPointCalculationMethod(raypointsmethod.withInterpolation)
            alRecons.setOutgoingRayAlgorithm(optionOutgoing)
            alRecons.setSubdivisionSelfAbsMat(2)
            alRecons.setI0(self.I0)
            alRecons.setRandSeedToZero(True)
            alRecons.setVoxelSize(1.0*metricsystem.cm)
            
            reconstructedPhantom = alRecons.iterate(2)
            _logger.info("testing outgoingMethod " + str(optionOutgoing))
            _logger.info("theoretical value = " + str(self.interactionProba))
            _logger.info("freeART value = " + str(reconstructedPhantom[self.ypos, self.xpos]) )

            np.testing.assert_allclose(initial_phantom, reconstructedPhantom, rtol=0.12 , atol=1e-2 )

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testRayParameters))
    
    return test_suite

if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Python utils for freeart."""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/05/2016"

import numpy


def computeDetectorPosition(detPos, angle):
    """
    Rotate the detector position along the origin of the given angle

    :param float angle: the angle of rotation (in rad)
    """
    vectorX = (0.0, 1.0, 0.0)
    detectorDis = numpy.sqrt(detPos[0] * detPos[0] + detPos[1] * detPos[1])
    currentAngle = getAngle(vectorX, detPos)
    rotAngle = currentAngle + angle
    return numpy.array((numpy.sin(rotAngle) * detectorDis,
                        numpy.cos(rotAngle) * detectorDis,
                        detPos[2]))


def getDistance(pt1, pt2):
    """
    Return the distance between two points
    """
    return numpy.sqrt((pt1[0]-pt2[0])*(pt1[0]-pt2[0]) + (pt1[1]-pt2[1])*(pt1[1]-pt2[1]) + (pt1[2]-pt2[2])*(pt1[2]-pt2[2]))


def getDistance_2D(pt1, pt2):
    """"
    Return the distance between two points
    """
    return numpy.sqrt((pt1[0]-pt2[0])*(pt1[0]-pt2[0]) + (pt1[1]-pt2[1])*(pt1[1]-pt2[1]))


def dotproduct(v1, v2):
    """
    Return the dot product of v1 qnd v2
    """
    return sum((a*b) for a, b in zip(v1, v2))


def length(v):
    """
    Return the length of the vector v
    """
    return numpy.sqrt(dotproduct(v, v))


def getAngle(v1, v2):
    """
    Return the angle between v1 and v2
    """
    return numpy.arccos(dotproduct(v1, v2) / (length(v1) * length(v2)))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""test that the TxConfig and FluoConfig class are correct
"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

import unittest
from freeart.configuration import config, structs, fileInfo
import numpy


class TestConfigurationClasses(unittest.TestCase):
    """Make sure the configuration classes are correctly instanciated"""
    def testBaseClass(self):
        """Test the Base class conditions : _ReconsConfig"""
        l = config._ReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                 projections='2:6',
                                 minAngle=0,
                                 maxAngle=numpy.pi)

        with self.assertRaises(ValueError):
            l = config._ReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                     projections='2:6',
                                     minAngle=0,
                                     maxAngle=3*numpy.pi)

        l = config._ReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                 projections='2:6',
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 I0=1)
        self.assertTrue(l.useAFileForI0 is False)

        l = config._ReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                 projections='2:6',
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 I0='dsadsda')
        self.assertTrue(l.useAFileForI0 is True)

    def testItRecons(self):
        "test _ItReconsConfig class"
        l = config._ItReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                 projections='2:6',
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 dampingFactor=1.,
                                 precision='double')
        self.assertTrue(l.reconsType == config._ReconsConfig.TX_ID)
        self.assertTrue(l.projections == '2:6')
        self.assertTrue(l.dampingFactor == 1.)
        self.assertTrue(l.precision == 'double')

    def testTxConfig(self):
        "test TxConfig class"
        l = config.TxConfig(sinograms=None,
                            projections='2:6',
                            minAngle=0,
                            maxAngle=numpy.pi,
                            dampingFactor=1.,
                            precision='simple')
        self.assertTrue(l.reconsType == config._ReconsConfig.TX_ID)
        self.assertTrue(l.projections == '2:6')
        self.assertTrue(l.dampingFactor == 1.)
        self.assertTrue(l.computeLog is True)
        self.assertTrue(l.precision == 'simple')
        l.addSinogram(structs.TxSinogram(name='sino1',
                                         data=numpy.zeros((10, 10))))
        l.addSinogram(structs.TxSinogram(name='sino2',
                                         data=numpy.zeros((10, 10))))
        self.assertTrue(len(l.sinograms) is 2)

    def testFluoConfig(self):
        """test FluoConfig class"""
        detector = structs.Detector(x=1, y=2, z=3, width=0.1)
        l = config.FluoConfig(outBeamCalMethod=None,
                              sinoI0=None,
                              absMat=None,
                              isAbsMatASinogram=True,
                              detector=detector,
                              minAngle=0,
                              maxAngle=numpy.pi,
                              dampingFactor=1.)
        self.assertTrue(l.reconsType == config._ReconsConfig.FLUO_ID)
        self.assertTrue(l.projections == None)
        self.assertTrue(l.dampingFactor == 1.)
        self.assertTrue(l.absMat == structs.AbsMatrix())
        self.assertTrue(l.detector == detector)


class TestConfigToDict(unittest.TestCase):
    """Make sure the toDict function of the reconstruction configuration are
    working"""
    def testBaseClass(self):
        """Test the Base class conditions : _ReconsConfig"""
        l = config._ReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                 projections='2:6',
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 includeLastProjection=True,
                                 oversampling=12)
        l.toDict()

    def testItRecons(self):
        "test _ItReconsConfig class"
        l = config._ItReconsConfig(reconsType=config._ReconsConfig.TX_ID,
                                   projections='2:6',
                                   minAngle=0,
                                   maxAngle=numpy.pi,
                                   dampingFactor=1.,
                                   solidAngleOff=True)
        l.toDict()

    def testTxConfig(self):
        "test TxConfig class"
        filePath = 'test.edf'
        sinograms = [
            structs.TxSinogram(
                fileInfo=fileInfo.MatrixFileInfo(file_path=filePath)),
            structs.TxSinogram(
                fileInfo=fileInfo.MatrixFileInfo(file_path=filePath))
        ]
        conf = config.TxConfig(sinograms=sinograms,
                               projections='2:6',
                               minAngle=0,
                               maxAngle=numpy.pi,
                               dampingFactor=1.)
        conversion = conf.toDict()
        self.assertTrue(conf == config.TxConfig()._fromDict(conversion))

    def testFluoConfig(self):
        """test FluoConfig class"""
        filePath = 'test.edf'
        sinoTest = structs.FluoSino(
            fileInfo=fileInfo.MatrixFileInfo(file_path=filePath),
            name='mySinogram',
            physElmt='O',
            ef=12,
            selfAbsMat=None)
        detector = structs.Detector(x=1, y=2, z=3, width=0.1)

        conf = config.FluoConfig(outBeamCalMethod=None,
                                 sinoI0=None
                                 ,
                                 absMat=None,
                                 isAbsMatASinogram=True,
                                 detector=detector,
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 dampingFactor=1.,
                                 centerOfRotation=6.,
                                 acquiInverted=True)
        conf.addSinogram(sinoTest)
        conversion = conf.toDict()
        self.assertTrue(conf == config.FluoConfig()._fromDict(conversion))

        conf = config.FluoConfig(outBeamCalMethod=None,
                                 sinoI0=None,
                                 absMat=None,
                                 isAbsMatASinogram=True,
                                 detector=detector,
                                 minAngle=0,
                                 maxAngle=numpy.pi,
                                 dampingFactor=1.,
                                 centerOfRotation=6.,
                                 acquiInverted=True,
                                 sinograms={filePath: sinoTest})
        conversion = conf.toDict()
        self.assertTrue(conf == config.FluoConfig()._fromDict(conversion))


def suite():
    test_suite = unittest.TestSuite()
    for t in (TestConfigurationClasses, TestConfigToDict):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(t))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/08/2017"

import unittest
from freeart import utils
import numpy

import logging
_logger = logging.getLogger("unitests")


class testUtils(unittest.TestCase):
    """
    Test some function of test_utils
    """
    DET_POS = numpy.array((1000, 0, 0))

    def testChangeDetectorPosition(self):
        self.assertTrue(
            numpy.allclose(utils.computeDetectorPosition(self.DET_POS,
                                                         0.0),
                           self.DET_POS))
        self.assertTrue(
            numpy.allclose(utils.computeDetectorPosition(self.DET_POS,
                                                         numpy.pi / 2.0),
                           (0, -1000, 0)))
        self.assertTrue(
            numpy.allclose(utils.computeDetectorPosition(self.DET_POS,
                                                         numpy.pi),
                           (-1000, 0, 0)))


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(testUtils))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/08/2017"


import logging
import os
import shutil
import tempfile
import unittest
import numpy as np
import h5py
import numpy
from freeart.utils import testutils
from freeart.configuration import config, structs, fileInfo
from freeart.interpreter.configinterpreter import FluoConfInterpreter
from freeart.interpreter.test import utils
from freeart.utils import reconstrutils

_logger = logging.getLogger("unitests")


class TestConfig(unittest.TestCase):
    """Test that Config functions are working"""

    def setUp(self):
        self.setValidConfigFiles()
        self.setUnvalidConfigFiles()

        # write valid files
        self.writeValidTx(self.configfile_valid_tx)
        self.writeFluoFileValid(self.configfile_valid_fluo_file)
        self.writeFluoFolderValid(self.configfile_valid_fluo_folder)

        # write unvalid files
        self.writeUnvalidTx(self.configfile_unvalid_1)
        self.writeUnvalidFluo2(self.configfile_unvalid_3)

    def setValidConfigFiles(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfile_valid_fluo_file = os.path.join(self.tempdir,
                                                       "tmp_config_valid_fluo_1.cfg")
        self.configfile_valid_fluo_folder = os.path.join(self.tempdir,
                                                         "tmp_config_valid_fluo_2.cfg")
        self.configfile_valid_tx = os.path.join(self.tempdir,
                                                "tmp_config_valid_tx.cfg")

    def setUnvalidConfigFiles(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfile_unvalid_1 = os.path.join(self.tempdir,
                                                 "tmp_config_unvalid_1.cfg")
        self.configfile_unvalid_3 = os.path.join(self.tempdir,
                                                 "tmp_config_unvalid_3.cfg")

    def tearDown(self):
        self.removeValidConfigFiles()
        self.removeUnvalidConfigFiles()

    def removeUnvalidConfigFiles(self):
        if os.path.isfile(self.configfile_unvalid_1):
            os.unlink(self.configfile_unvalid_1)

    def removeValidConfigFiles(self):
        if os.path.isfile(self.configfile_valid_fluo_file):
            os.unlink(self.configfile_valid_fluo_file)

        if os.path.isfile(self.configfile_valid_fluo_folder):
            os.unlink(self.configfile_valid_fluo_folder)

        if os.path.isfile(self.configfile_valid_tx):
            os.unlink(self.configfile_valid_tx)


class TestDatasetToH5(unittest.TestCase):
    """
    class to test the validity of `setFileInfoToH5File` function in the 
    _ReconsConfig
    """
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        self.confFluo = testutils.createConfigFluo(self.tempdir,
                                                  fileExtension='.edf',
                                                  withMaterials=True,
                                                  addAbsMat=True,
                                                  fileExtensionMat='.npy',
                                                  addSelfAbsMat=True)
        self.confTx = testutils.createConfigTx(self.tempdir)
        i0 = structs.I0Sinogram(fileInfo=fileInfo.EDFMatrixFileInfo(os.path.join(self.tempdir, 'io.edf')),
                                data=numpy.zeros(100).reshape(10, 10))
        self.confTx.I0 = self.confFluo.I0 = i0
        self.h5File = os.path.join(self.tempdir, 'configFile.h5')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testTxNotSaving(self):
        self.confTx.setFileInfoToH5File(refFile=self.h5File, save=False)
        self.assertFalse(os.path.isfile(self.h5File))

    def testTxSaving(self):
        self.confTx.setFileInfoToH5File(refFile=self.h5File, save=True)
        file = h5py.File(self.h5File, mode='r')
        self.assertTrue(structs.TxSinogram.TX_SINOGRAM_DATASET.strip('::') in file)
        self.assertTrue(structs.I0Sinogram.I0_DATASET.strip('::') in file)

    def testFluoSaving(self):
        self.confFluo.setFileInfoToH5File(refFile=self.h5File, save=False)
        self.assertFalse(os.path.isfile(self.h5File))

    def testFluoNotSaving(self):
        self.confFluo.setFileInfoToH5File(refFile=self.h5File, save=True)
        file = h5py.File(self.h5File, mode='r')
        self.assertTrue(structs.I0Sinogram.I0_DATASET.strip('::') in file)
        self.assertTrue(structs.MaterialsDic.MATERIALS_DICT.strip('::') in file)
        self.assertTrue(structs.MatComposition.MAT_COMP_DATASET.strip('::') in file)
        self.assertTrue(structs.AbsMatrix.ABS_MAT_INDEX.strip('::') in file)
        self.assertTrue((structs.FluoSino.SINO_DATASET.strip('::') + '1') in file)
        self.assertTrue((structs.SelfAbsMatrix.MATERIALS_DICT.strip('::') + '1') in file)

class TestFluoCFGFile(unittest.TestCase):
    """
    Test that some cfg file are correctly parsed and interpreted
    """

    MATRIX_DIM = 10

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configFile = os.path.join(self.tempdir + 'conf.cfg')
        self.absFile = os.path.join(self.tempdir + 'absMatFile.edf')
        self.selfAbsFile = os.path.join(self.tempdir + 'selfAbsMatFile.edf')
        reconstrutils.saveMatrix(
            np.random.rand(self.MATRIX_DIM, self.MATRIX_DIM),
            self.absFile
        )

        reconstrutils.saveMatrix(
            np.random.rand(self.MATRIX_DIM, self.MATRIX_DIM),
            self.selfAbsFile
        )

        utils.writeGeneralSettings(f=self.configFile,
                                   reconsType=config._ReconsConfig.FLUO_ID)
        utils.writeFluoGeneral(f=self.configFile,
                               absFileIsASino=False,
                               absFile=self.absFile,
                               sampComposition=None,
                               materialsFile=None)

        utils.addSinoFile(f=self.configFile,
                          dim=self.MATRIX_DIM,
                          nToAdd=2,
                          physElmts=['H', 'O'],
                          selfAbsPath=self.selfAbsFile,
                          tempdir=self.tempdir)

        utils.addNormalization(f=self.configFile,
                               rotCenter=5,
                               normI0FrmFile=False,
                               i0=1.0,
                               computeMinusLog=True)

        utils.addReconsProperties(f=self.configFile,
                                  voxelSize=1.0,
                                  oversampling=2,
                                  relaxationFactor=0.23)

        utils.addReductionData(f=self.configFile)
        utils.addProjInfo(f=self.configFile,
                          startProj=0,
                          endProj=self.MATRIX_DIM-1)
        utils.addDetectorInfo(f=self.configFile,
                              detWidth=10,
                              detX=10.0,
                              detY=0.0,
                              detZ=0.0)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testFluo(self):
        interpreter = FluoConfInterpreter(self.configFile)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestConfig, TestFluoCFGFile):
        test_suite.addTest(
            unittest.defaultTestLoader.loadTestsFromTestCase(ui))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



  #!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Simple application doing data.max() - data
"""

__authors__ = ["H> Payno"]
__license__ = "MIT"
__date__ = "08/08/2018"


from freeart.utils import (reconstrutils, outgoingrayalgorithm, raypointsmethod)
from freeart.unitsystem import metricsystem
import argparse
import sys
import logging
import numpy


def getinputinfo():
    return "freeart createfluosino sample.edf absmat.edf selfasbmat.edf [nangles]"



def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('phantomfile',
                        help='edf file containing the sample density to generate the fluorescence sinogram')
    parser.add_argument('absmatfile',
                        help='definition of the sample absorption (in g.cm-2)')
    parser.add_argument('selfabsmatfile',
                        help='definition of the sample self attenuation (in g.cm-2)')
    parser.add_argument('--output', '-o',
                        dest="output",
                        default=None,
                        help='outputfile')
    parser.add_argument('--I0', '--i0',
                        dest="i0",
                        type=float,
                        default=1.0)
    parser.add_argument('--voxelsize',
                        dest="voxelsize",
                        type=float,
                        default=1.0*metricsystem.centimeter)
    parser.add_argument('--nangle',
                        dest="nangle",
                        type=int,
                        default=None,
                        help='number of projections')
    parser.add_argument('--center',
                        dest="center",
                        type=int,
                        default=None,
                        help='change the center once the sinogram has been created. Can only be an integer for now')
    parser.add_argument('--halfacq',
                        dest="halfacq",
                        action="store_true",
                        default=False,
                        help='Do we want to make a sinogram from 0 to 180 or from 0 to 360 degree')
    parser.add_argument('--debug',
                        dest="debug",
                        action="store_true",
                        default=False,
                        help='Set logging system in debug mode')
    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    _halfacq = options.halfacq
    _nangle = options.nangle
    if _nangle is None:
        _nangle = 180 if _halfacq is True else 360

    _output = options.output
    if _output and _output.endswith('.edf') is False:
        _output = _output + '.edf'

    phantom_file = options.phantomfile
    assert phantom_file.lower().endswith('.edf')
    absmat_file = options.absmatfile
    assert absmat_file.lower().endswith('.edf')
    selfabsmat_file = options.selfabsmatfile
    assert selfabsmat_file.lower().endswith('.edf')

    phantom_data = reconstrutils.LoadEdf_2D(phantom_file)
    phantom_data = reconstrutils.increaseMatSize(phantom_data, False)

    absmat_data = reconstrutils.LoadEdf_2D(absmat_file)
    absmat_data = reconstrutils.increaseMatSize(absmat_data, False)

    selfabsmat_data = reconstrutils.LoadEdf_2D(selfabsmat_file)
    selfabsmat_data = reconstrutils.increaseMatSize(selfabsmat_data, False)

    detPos = (1000*metricsystem.cm, 0, 0) # 90 degree detector
    detectorRadius = 1*metricsystem.cm
    detSetup = [(detPos, detectorRadius)]
    oversampling = 2
    
    print('----- configuration ---')
    print('phantom file: ', phantom_file)
    print('absorption matrix file: ', absmat_file)
    print('self absorption mattrix file: ', selfabsmat_file)
    print('half acq: ', _halfacq)
    print('oversampling: ', oversampling)
    print('I0: ', options.i0)
    if options.center is not None:
        print('center: ', int(options.center))
    print('-----------------------')
    print('start sinogram generation')

    # TODO: add possibility to define the voxel size in the command line

    # run the sinogram generation
    sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(phantom=phantom_data,
                                                             absMat=absmat_data,
                                                             selfAbsMat=selfabsmat_data,
                                                             numAngle=_nangle,
                                                             minAngle=0.0,
                                                             maxAngle=numpy.pi if _halfacq else numpy.pi * 2.0,
                                                             turnOffSolidAngle=True,
                                                             oversampling=oversampling,
                                                             beamCalcMeth=raypointsmethod.withInterpolation,
                                                             outRayPtCalcMeth=outgoingrayalgorithm.rawApproximation,
                                                             voxelSize=options.voxelsize,
                                                             detSetup=detSetup,
                                                             I0=options.i0)
    sinogram = reconstrutils.decreaseMatSize(sinogram)
    if options.center:
        center = int(options.center)
        res = numpy.zeros((sinogram.shape[0], center * 2))
        if center * 2.0 > sinogram.shape[1]:
            res[:, center * 2 - sinogram.shape[1]:] = sinogram
        else:
            res = sinogram[:, sinogram.shape[1] / 2 - center]
        sinogram = res

    if _output is not None:
        reconstrutils.saveMatrix(data=sinogram, fileName=_output, overwrite=True)
        # TODO: else plot it using silx
    print('end sinogram generation')

if __name__ == "__main__":
    """
    """
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2017"


import unittest
import numpy
import silx
if silx._version.MINOR < 7:
    from silx.test.utils import ParametricTestCase
else:
    from silx.utils.testutils import ParametricTestCase
from freeart.unitsystem import metricsystem
from freeart.utils import testutils, reconstrutils, raypointsmethod, outgoingrayalgorithm
import logging

_logger = logging.getLogger("unitests")


class TestFluoIncomingBeam_2D(ParametricTestCase):
    """
    Here we want to test the incoming beam attenuation in a stacks of voxels.
    """
    origin = (0.0, 0.0)
    detectorRadius = 10.0*metricsystem.cm
    detectorArea = detectorRadius * 2.0
    detectorWidth = detectorRadius*2.0
    detectorDistance = 1000*metricsystem.cm

    oversampling = 40
    phDimX = 31
    phDimY = 31
    phDimZ = 1

    xpos = 15
    ypos = 15
    zpos = 0

    interactionProba = 1.0
    outgoingAbsorptionFactor = 0.0  # g.cm2

    voxelPosInPhantomRef = (xpos, ypos, zpos)
    voxelPosInExpRef = testutils.toExperimentationRef(voxelPosInPhantomRef,
                                                      phDimX,
                                                      phDimY,
                                                      phDimZ)

    def testProjectorIncomingBeamAbsorption(self):
        """
        We have n lines of material with a particular attenuation (incoming
        beam attenuation).
        The outgoing beam will not be attenuated>
        Then for different incoming beam angle we will compare the values
        returned by FreeART and given by the theory.
        """
        
        widthsToTest = [5]  # in voxels
        incomingAbsorptionToTest = [1e-4, 2.0]  # in g/cm2

        methodGBeams = raypointsmethod.withInterpolation
        optionOutgoing = outgoingrayalgorithm.rawApproximation


        detAngle = 0
        detPos = (numpy.sin(detAngle)*self.detectorDistance,
                  numpy.cos(detAngle)*self.detectorDistance,
                  self.zpos)
        detSetup = [(detPos, self.detectorRadius)]

        for precision in (numpy.float32, numpy.float64):
            for incomingAbsorptionFactor in incomingAbsorptionToTest:
                for width in widthsToTest:
                    for voxelSize in [0.2*metricsystem.cm, 0.3*metricsystem.mm]:
                        with self.subTest(incomingAbsorptionFactor=incomingAbsorptionFactor,
                                          width=width, voxelSize=voxelSize,):

                            phantom = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                  dtype=precision)
                            absMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                 dtype=precision)
                            selfAbsMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                     dtype=precision)

                            for delta in numpy.arange(0, width, 1) :
                                phantom[self.ypos+delta, :, self.zpos] = self.interactionProba
                                absMat[self.ypos+delta, :, self.zpos] = incomingAbsorptionFactor
                            sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(phantom,
                                                                                     absMat,
                                                                                     selfAbsMat,
                                                                                     360/15 +1,
                                                                                     detSetup,
                                                                                     self.oversampling,
                                                                                     methodGBeams,
                                                                                     optionOutgoing,
                                                                                     voxelSize=voxelSize)
                            I_FreeARTValue_0 = sinogram[0][0][self.xpos]
                            I_FreeARTValue_30 = sinogram[0][2][self.xpos]
                            I_FreeARTValue_45 = sinogram[0][3][self.xpos]
                            I_FreeARTValue_60 = sinogram[0][4][self.xpos]
                            I_FreeARTValue_180 = sinogram[0][12][self.xpos]

                            I_theoreticalValue_0 = testutils.getFluoTheoreticalValue(
                                detPos,
                                self.detectorRadius,
                                self.voxelPosInExpRef,           # < voxel position in the freeART reference
                                incomingAbsorptionFactor,        # < mass attenuation for the incoming beam
                                self.outgoingAbsorptionFactor,   # < mass attenuation for the outgoing beam
                                numpy.pi/2.0,                    # < angle of entry of the incoming beam
                                numpy.pi/2.0 - detAngle,         # < outgoing beam angle to the sample
                                1.0,                             # < tau
                                width*voxelSize                  # < When sampling value is 1, then we will get the attenuation on the whole voxel.
                                )

                            I_theoreticalValue_30 = testutils.getFluoTheoreticalValue(
                                detPos,
                                self.detectorRadius,
                                self.voxelPosInExpRef,            # < voxel position in the freeART reference
                                incomingAbsorptionFactor,         # < mass attenuation for the incoming beam
                                self.outgoingAbsorptionFactor,    # < mass attenuation for the outgoing beam
                                numpy.pi/2.0 - 30.0*numpy.pi/180, # < angle of entry of the incoming beam
                                numpy.pi/2.0 - detAngle,          # < outgoing beam angle to the sample
                                1.0,
                                width*voxelSize
                                )

                            I_theoreticalValue_45 = testutils.getFluoTheoreticalValue(
                                detPos,
                                self.detectorRadius,
                                self.voxelPosInExpRef,               # < voxel position in the freeART reference
                                incomingAbsorptionFactor,            # < mass attenuation for the incoming beam
                                self.outgoingAbsorptionFactor,       # < mass attenuation for the outgoing beam
                                numpy.pi/2.0 - 45.0*numpy.pi/180.0,  # < angle of entry of the incoming beam
                                numpy.pi/2.0 - detAngle,             # < outgoing beam angle to the sample
                                1.0,
                                width*voxelSize
                                )

                            I_theoreticalValue_60 = testutils.getFluoTheoreticalValue(
                                detPos,
                                self.detectorRadius,
                                self.voxelPosInExpRef,              # < voxel position in the freeART reference
                                incomingAbsorptionFactor,           # < mass attenuation for the incoming beam
                                self.outgoingAbsorptionFactor,      # < mass attenuation for the outgoing beam
                                numpy.pi/2.0 - 60.0*numpy.pi/180.0, # < angle of entry of the incoming beam
                                numpy.pi/2.0 - detAngle,            # < outgoing beam angle to the sample
                                1.0,
                                width*voxelSize
                                )

                            I_theoreticalValue_180 = testutils.getFluoTheoreticalValueForTransmission(
                                detPos,
                                self.detectorRadius,
                                self.voxelPosInExpRef,          # < voxel position in the freeART reference
                                incomingAbsorptionFactor,       # < mass attenuation for the incoming beam
                                self.outgoingAbsorptionFactor,  # < mass attenuation for the outgoing beam
                                numpy.pi/2.0,                   # < angle of entry of the incoming beam
                                numpy.pi/2.0 - detAngle,        # < outgoing beam angle to the sample
                                1.0,
                                width*voxelSize
                                )

                        self.assertAlmostEqual(I_theoreticalValue_0, I_FreeARTValue_0, 5)
                        self.assertAlmostEqual(I_theoreticalValue_30, I_FreeARTValue_30, 5)
                        self.assertAlmostEqual(I_theoreticalValue_45, I_FreeARTValue_45, 5)
                        self.assertAlmostEqual(I_theoreticalValue_60, I_FreeARTValue_60, 5)
                        self.assertAlmostEqual(I_theoreticalValue_180, I_FreeARTValue_180, 5)


class TestFluoOutgoingBeam_2D(ParametricTestCase):
    """the unit test related to the fluorescence projector, focus
    on the incoming beam attenuations.
    """
    # detector properties
    detectorRadius = 10.0*metricsystem.cm
    detectorArea = detectorRadius * 2.0
    detectorWidth = detectorRadius * 2.0
    detectorDistance = 1000*metricsystem.cm
    # Dimensions of the phantom to test
    phDimX = 21
    phDimY = 21
    phDimZ = 1

    xpos = 10
    ypos = 10
    zpos = 0

    origin = (0.0, 0.0)
    voxelPosInPhantomRef = (xpos, ypos, zpos)
    voxelPosInExpRef = testutils.toExperimentationRef(voxelPosInPhantomRef,
                                                      phDimX,
                                                      phDimY,
                                                      phDimZ)

    # we are setting an hight value to get close from the theoretical values
    # we need a large number of ray especially for the case where we are not
    # using the interpolation process.
    oversampling = 40
    # For this test case we are skipping the attenuation of the incoming beam
    # and only focusing on the outgoing
    interactionProba = 1.0
    incomingAbsorptionFactor = 0.0

    def test_outgoing_beam_absorption_width_voxels_y(self):
        """Test the attenuation for the outgoing beam hitting n lines of
        material.
        Material attenuation of the incoming beam is zero.
        Material attenuation of the outgoing beam is outgoingAbsorptionFactor
        To make sure results are ok we are testing different detector position,
        which are directly defining the angle of the outgoing beam.
        """
        width = 8
        optionOutgoing = outgoingrayalgorithm.createOneRayPerSamplePoint

        for precision in (numpy.float32, numpy.float64):
            for methodGBeams in [raypointsmethod.withInterpolation, raypointsmethod.withoutInterpolation]:
                for outgoingAbsorptionFactor in (1.e-5, 2.0):
                    for detAngle in (0, 60*numpy.pi/180):
                        detPos = (numpy.sin(detAngle) * self.detectorDistance,
                                  numpy.cos(detAngle) * self.detectorDistance,
                                  self.zpos)
                        detSetup = [(detPos, self.detectorRadius)]

                        for voxelSize in (0.2 * metricsystem.cm, 0.3 * metricsystem.nm):

                            with self.subTest(precision=precision, calculationBMethod=methodGBeams,
                                              outgoingAbsorptionFactor=outgoingAbsorptionFactor,
                                              voxelSize=voxelSize):

                                ### RUN the FREE ART EXPERIMENTATION
                                # create the simple phantom
                                phantom = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                      dtype=precision)
                                absMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                     dtype=precision)
                                selfAbsMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                         dtype=precision)

                                for delta in numpy.arange(0, width, 1):
                                    phantom[self.ypos+delta, :, self.zpos] = self.interactionProba
                                    selfAbsMat[self.ypos+delta, :, self.zpos] = outgoingAbsorptionFactor

                                sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(phantom=phantom,
                                                                                         absMat=absMat,
                                                                                         selfAbsMat=selfAbsMat,
                                                                                         numAngle=2,
                                                                                         detSetup=detSetup,
                                                                                         oversampling=self.oversampling,
                                                                                         beamCalcMeth=methodGBeams,
                                                                                         outRayPtCalcMeth=optionOutgoing,
                                                                                         voxelSize=voxelSize)

                                # The momentum when we are going trought the n voxel ( n == width )
                                I_FreeARTValue_NVoxels = sinogram[0][0][self.xpos]

                                # GET THE THEORETICAL VALUES
                                I_theoreticalValue_NVoxels = testutils.getFluoTheoreticalValue(
                                    detPos,
                                    self.detectorRadius,
                                    self.voxelPosInExpRef,           # < voxel position in the freeART reference
                                    self.incomingAbsorptionFactor,   # < mass attenuation for the incoming beam
                                    outgoingAbsorptionFactor,        # < mass attenuation for the outgoing beam
                                    numpy.pi/2.0,                    # < angle of entry of the incoming beam
                                    numpy.pi/2.0 - detAngle,         # < outgoing beam angle to the sample
                                    1.0,                             # < tau
                                    width*voxelSize                  # < material width
                                    )
                                places = 6 if precision is numpy.float64 else 3
                                self.assertAlmostEqual(I_theoreticalValue_NVoxels,
                                                       I_FreeARTValue_NVoxels,
                                                       places)


class TestFluoComplexCases(ParametricTestCase):
    """the unit test related to the compton projector and debug"""

    # Dimensions of the phantom to test
    phDimX = 41
    phDimY = 41
    phDimZ = 1

    origin = (0.0, 0.0)

    xpos = 20
    ypos = 20
    zpos = 0

    voxelPosInPhantomRef = (xpos, ypos, zpos)
    voxelPosInExpRef = testutils.toExperimentationRef(voxelPosInPhantomRef,
                                                      phDimX,
                                                      phDimY,
                                                      phDimZ)

    # detector properties
    detectorRadius = 10.0*metricsystem.cm
    detectorArea = detectorRadius * 2.0
    detectorWidth = detectorRadius * 2.0

    # we are setting an hight value to get close from the theoretical values
    oversampling = 20
    # For this test case we are skipping the attenuation of the incoming beam and only focusing on the outgoing
    interactionProba = 1.0
    incomingAbsorptionFactor = 0.0
    # The width of the stack to test to make sure the attenuation is well propagated
    width = 5  # nb voxel.
    # The outgoing beam attenuation to consider for this test case
    outgoingAbsorptionFactor = 1.0
    detectorDistance = 1000

    def test_outgoing_beam_absorption_width_voxels_y(self):
        """
        This is the cqse where the incoming ray is hitting a voxel with only an
        interaction probability to 1. At the time the beam haven t been
        attenuated by any material.
        Then the outgoing beam is generated (at xpos, ypos) and going to the
        detector.
        On this path the ray will cross an area with an attenuation of
        outgoingAbsorptionFactor. Throught a distance of
        sqrt(width*width + width*width).
        This test is only verifying that this attenuation for this cqse is well
        taking into account.
        """
        detAngle = numpy.pi / 4.0  # 45 degree detector
        # we are in the trigonometric reference
        detPos = (numpy.sin(detAngle)*self.detectorDistance,
                  numpy.cos(detAngle)*self.detectorDistance,
                  self.zpos)
        detSetup = [(detPos, self.detectorRadius)]

        # RUN the FREE ART EXPERIMENTATION
        # create the simple phantom
        phantom = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                              dtype=numpy.float64)
        absMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                             dtype=numpy.float64)
        selfAbsMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                 dtype=numpy.float64)

        phantom[self.ypos, self.xpos, self.zpos] = self.interactionProba

        for delta in numpy.arange(0, self.width, 1):
            selfAbsMat[self.ypos + delta + 2, self.xpos+1:, self.zpos] = self.outgoingAbsorptionFactor

        # Define the option to compute the attenatuion of rays
        methodGBeams = raypointsmethod.withInterpolation
        optionOutgoing = outgoingrayalgorithm.createOneRayPerSamplePoint

        sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
            phantom, absMat, selfAbsMat, 5, detSetup, self.oversampling, methodGBeams, optionOutgoing)

        # The momentum when we are going trought the n voxel ( n == width )
        I_FreeARTValue_NVoxels = sinogram[0][0][self.xpos]

        # width : because we are in the 45 degree case
        distInMaterialForTheOutgoingBeam = numpy.sqrt(self.width*self.width*2)
        I_theoreticalValue_NVoxels = numpy.exp(-distInMaterialForTheOutgoingBeam*self.outgoingAbsorptionFactor)*testutils.getSolidAngle(self.voxelPosInExpRef, detPos, self.detectorRadius)

        self.assertAlmostEqual(I_theoreticalValue_NVoxels, I_FreeARTValue_NVoxels, 5)


class TestFluoIncomingAndOutgoingBeam(ParametricTestCase):
    """
    the unit test related to the compton projector and constructor
    """

    # detector properties
    detectorRadius = 10.0*metricsystem.cm
    detectorArea = detectorRadius * 2.0
    detectorWidth = detectorRadius * 2.0
    detectorDistance = 1000*metricsystem.cm
    # Dimensions of the phantom to test
    phDimX = 31  # voxel
    phDimY = 31  # voxel
    phDimZ = 1  # voxel

    xpos = 15  # voxel
    ypos = 15  # voxel
    zpos = 0   # voxel

    origin = (0.0, 0.0, 0.0)

    voxelPosInPhantomRef = (xpos, ypos, zpos)
    voxelPosInExpRef = testutils.toExperimentationRef(voxelPosInPhantomRef, phDimX, phDimY, phDimZ)

    # we are setting an hight value to get close from the theoretical values
    oversampling = 20
    interactionProba = 1.0

    def test_absorption_in_a_stack(self):
        """
        Test the attenuation for the outgoing beam.
        with have n lines of material (n == width ) where we are testing over
        different absorption.
        for the incoming beam and for the outgoing beam.
        Then we compare result given by freeART and by the theory
        """
        width = 6

        methodGBeams = raypointsmethod.withInterpolation
        optionOutgoing = outgoingrayalgorithm.createOneRayPerSamplePoint

        for precision in (numpy.float32, numpy.float64):
            for inAbsorbFact in (1.e-5, 2.0):
                for outAbsorbFact in (1.e-5, 2.0):
                    for detAngle in (0, 30.0*numpy.pi/180.0):
                        # we are in the trigonometric reference
                        detPos = (numpy.sin(detAngle)*self.detectorDistance,
                                  numpy.cos(detAngle)*self.detectorDistance,
                                  self.zpos)
                        detSetup = [(detPos, self.detectorRadius)]

                        for voxelSize in [1.0*metricsystem.mm, 1.0*metricsystem.cm]:
                            with self.subTest(precision=precision, incoAbsMat=inAbsorbFact,
                                              outAbsorbFact=outAbsorbFact, detAngle=detAngle,
                                              voxelSize=voxelSize):
                                # create the simple phantom
                                phantom = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                      dtype=precision)
                                absMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                     dtype=precision)
                                selfAbsMat = numpy.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                         dtype=precision)

                                # dans ce cas dans le repere de freeART les lignes se trouve en (:, -5:-7)
                                for delta in numpy.arange(0, width, 1):
                                    phantom[self.ypos+delta, :, self.zpos] = self.interactionProba
                                    selfAbsMat[self.ypos+delta, :, self.zpos] = outAbsorbFact
                                    absMat[self.ypos+delta, :, self.zpos] = inAbsorbFact

                                sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(phantom=phantom,
                                                                                         absMat=absMat,
                                                                                         selfAbsMat=selfAbsMat,
                                                                                         numAngle=5,
                                                                                         detSetup=detSetup,
                                                                                         oversampling=self.oversampling,
                                                                                         beamCalcMeth=methodGBeams,
                                                                                         outRayPtCalcMeth=optionOutgoing,
                                                                                         voxelSize=voxelSize)

                                # The momentum when we are going trought the n voxel ( n == width )
                                I_FreeARTValue_NVoxels = sinogram[0][0][self.xpos]

                                ### GET THE THEORETICAL VALUES
                                I_theoreticalValue_NVoxels = testutils.getFluoTheoreticalValue(
                                    detPos=detPos,
                                    detectorRadius=self.detectorRadius,
                                    voxelPosInFreeARTRef=self.voxelPosInExpRef,
                                    mu_E0=inAbsorbFact,
                                    mu_EF=outAbsorbFact,
                                    psiIncoming=numpy.pi/2.0,
                                    psiOutgoing=numpy.pi/2.0 - detAngle,
                                    tau_E0=1.0,
                                    distInMat=width*voxelSize
                                )
                                places = 3 if precision is numpy.float32 else 5
                                self.assertAlmostEqual(I_theoreticalValue_NVoxels, I_FreeARTValue_NVoxels, places)


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFluoIncomingBeam_2D))
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFluoOutgoingBeam_2D))
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFluoComplexCases))
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFluoIncomingAndOutgoingBeam))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""

"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

from . import configIO
from . import config
import logging
from silx.io import dictdump

logger = logging.getLogger(__name__)


class IniConfigWriter(configIO.ConfigWriter):
    def write(self, filePath, reconsConfiguration, overwrite=False,
              merge=False):
        assert isinstance(reconsConfiguration, config._ReconsConfig)
        if merge is True:
            logger.warning('the `includeDatasetInFile` option is not managed'
                           'by the IniConfigWriter')
        dictdump.dicttoini(ddict=reconsConfiguration.toDict(),
                           inifile=filePath,
                           mode="w" if overwrite is True else 'a')


class IniConfigReader(configIO.ConfigReader):

    def read(self, filePath):
        dict = dictdump.load(filePath)
        if 'general_settings' not in dict:
            logger.warning('impossible to read reconstruction configuration,'
                           '`general_settings` not defined.'
                           'File structure is not valid')
            return None

        if 'reconstruction_type' not in dict['general_settings']:
            logger.warning(
                'impossible to read reconstruction configuration,'
                '`general_settings/reconstruction_type` not defined.'
                'File structure is not valid')
            return None

        reconsType = dict['general_settings']['reconstruction_type']
        if type(reconsType) is not str:
            reconsType = str(reconsType.tostring().decode())
        if reconsType not in self.dictReconsTypeToReconsConf:
            logger.error(
                'reconstruction type requested is not defined (%s)' % reconsType)
            return None
        conf = self.dictReconsTypeToReconsConf[reconsType]()._fromDict(dict)

        self._reloadDatasets(config=conf)
        return conf



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Contains some structures used for the reconstruction configuration
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

import numpy
import logging
from silx.io import configdict
from .fileInfo import MatrixFileInfo, FreeARTFileInfo, DictInfo
from freeart.utils import h5utils
from freeart.utils.reconstrutils import saveMatrix
from silx.io import dictdump

logger = logging.getLogger(__name__)


class DataStored(object):
    """
    Define a simple sinogram for his data and his storage information

    :param str h5defaultPath: the default location of this data inside a h5
                              file.
    :param name: possible tag of the data.
    :param fileInfo: information about file contained this data
    :param data: the data
    """
    def __init__(self, h5defaultPath=None, name=None, fileInfo=None, data=None):
        assert(fileInfo is None or isinstance(fileInfo, FreeARTFileInfo))
        self.h5defaultPath = h5defaultPath
        self.name = name
        self.fileInfo = fileInfo
        self.data = data

    def __eq__(self, other):
        if not isinstance(other, DataStored):
            return False
        dataEq = numpy.all(self.data == other.data) or \
                 (self.data is None and str(other.data) == '') or \
                 (other.data is None and str(self.data) == '')
        return self.name == other.name and \
               FreeARTFileInfo._equalOrNone(self.fileInfo, other.fileInfo) and \
               dataEq

    def __str__(self):
        return "name is %s " % self.name + \
               "fileInfo is %s " % self.fileInfo + \
               "data is %s " % self.data

    def loadData(self):
        if self.fileInfo is None or self.fileInfo.file_path() is None:
            logger.debug('No file info defined, can\'t load data')
            return self.data
        else:
            assert isinstance(self.fileInfo, FreeARTFileInfo)
            if self.data is not None:
                logger.debug('existing data, will be overwrite')
            data = self.fileInfo.load()
            if data is not None:
                self.data = data
            return self.data

    def save(self):
        # assert self.fileInfo is not None
        if self.fileInfo is None:
            logger.error('No file information specified, can\'t save the data')
            return
        return self.saveTo(self.fileInfo.file_path())

    def saveTo(self, path):
        if self.data is None:
            return

        self.fileInfo = MatrixFileInfo(file_path=path,
                                       data_path=self.fileInfo.data_path())
        extension = self.fileInfo.file_path().split('.')[-1].lower()
        if extension in ('h5', 'hdf5', 'nx', 'nxs'):
            h5utils.createH5WithDataSet(url=self.fileInfo,
                                        data=self.data,
                                        mode='a')
        elif extension in ('edf', ):
            saveMatrix(data=self.data, fileName=self.fileInfo.file_path(),
                       overwrite=True)
        elif extension in ('npy', ):
            numpy.save(self.fileInfo.file_path(), self.data)
        else:
            raise ValueError('extension type (%s) not managed' % extension)


indexSino = 0


class Sinogram(DataStored):
    """
    Sinogram class
    """
    SINO_DATASET = 'data/sinogram'
    def __init__(self, h5defaultPath=None, name=None, fileInfo=None, data=None):
        _name = name
        if h5defaultPath is None:
            global indexSino
            default = self.SINO_DATASET + str(indexSino)
            if _name is None:
                _name = 'sinogram' + str(indexSino)
            indexSino += 1
        else:
            default = h5defaultPath
        if _name is None:
            _name = 'sinogram' + str(indexSino)
            indexSino += 1
        DataStored.__init__(self, h5defaultPath=default,
                            name=_name, fileInfo=fileInfo, data=data)


class I0Sinogram(Sinogram):
    """I0 sinogram"""
    I0_DATASET = "data/I0"
    """Default location of the I0 sinogram in the .h5 file"""

    def __init__(self, fileInfo=None, data=None):
        Sinogram.__init__(self, h5defaultPath=self.I0_DATASET,
                          name='I0', fileInfo=fileInfo, data=data)


indexTxSino = 0


class TxSinogram(Sinogram):
    """Transmission sinogram"""
    TX_SINOGRAM_DATASET = "data/sinogramTX"
    """Default location of the transmission sinogram in the .h5 file"""

    def __init__(self, name=None, fileInfo=None, data=None):
        global indexTxSino
        Sinogram.__init__(self, h5defaultPath=self.TX_SINOGRAM_DATASET + str(indexTxSino),
                          name=name, fileInfo=fileInfo, data=data)
        indexTxSino += 1


indexFluoSino = 0


class FluoSino(Sinogram):
    """Fluorescence sinogram"""
    FLUO_DATASET = 'data/fluo_sinogram'
    """Default location of the fluorescence sinogram in the .h5 file"""

    def __init__(self, name, fileInfo, physElmt, ef, selfAbsMat, data=None):
        global indexFluoSino
        Sinogram.__init__(self, h5defaultPath=self.FLUO_DATASET + str(indexFluoSino),
                          name=name, fileInfo=fileInfo, data=data)
        indexFluoSino += 1
        self.physElmt = physElmt
        self.EF = ef
        assert(selfAbsMat is None or isinstance(selfAbsMat, SelfAbsMatrix))
        self.selfAbsMat = selfAbsMat

    def __eq__(self, other):
        return Sinogram.__eq__(self, other) and \
               self.EF == other.EF and \
               self.physElmt == other.physElmt

    def __str__(self):
        l = 'name = %s\n' % self.name
        l += 'physElmt = %s\n' % self.physElmt
        l += 'EF = %s\n' % self.EF
        l += str(self.selfAbsMat)
        return l


indexAbsMatrix = 0
class AbsMatrix(DataStored):
    """Absorption matrix class"""
    ABS_MAT_INDEX = "data/absmatrix/absm"
    """Default location of the absorption matrix in the .h5 file"""

    def __init__(self, name=None, fileInfo=None, data=None):
        global indexAbsMatrix
        DataStored.__init__(self, h5defaultPath=self.ABS_MAT_INDEX + str(indexAbsMatrix),
                            name=name, fileInfo=fileInfo, data=data)
        indexAbsMatrix += 1


indexSelfabsMatrix = 0
class SelfAbsMatrix(AbsMatrix):
    """Self absorption matrix class"""
    SELF_ABS_MAT_INDEX = "data/selfabsmatrix/selfabsm"
    """Default location of the self absorption matrix in the .h5 file"""

    def __init__(self, name=None, fileInfo=None, data=None):
        global indexSelfabsMatrix
        DataStored.__init__(self, h5defaultPath=self.SELF_ABS_MAT_INDEX + str(indexSelfabsMatrix),
                            name=name, fileInfo=fileInfo, data=data)
        indexSelfabsMatrix += 1


class MatComposition(DataStored):
    """Material /sample composition class"""
    MAT_COMP_DATASET = "data/materials/composition"
    """Default location of the sample composition in the .h5 file"""

    def __init__(self, name=None, fileInfo=None, data=None):
        DataStored.__init__(self, h5defaultPath=self.MAT_COMP_DATASET,
                            name=name, fileInfo=fileInfo, data=data)


class MaterialsDic(DataStored, configdict.ConfigDict):
    """Materials dictionary"""
    MATERIALS_DICT = "data/materials/matDict"
    """Default location of the materials dictionary in the .h5 file"""

    def __init__(self, name=None, fileInfo=None, data=None):
        configdict.ConfigDict.__init__(self, defaultdict=data)
        DataStored.__init__(self, h5defaultPath=self.MATERIALS_DICT,
                            name=name, fileInfo=fileInfo, data=data)

    def saveTo(self, path):
        self.fileInfo = DictInfo(file_path=path,
                                 data_path=self.fileInfo.data_path())
        assert self.fileInfo.is_valid()
        extension = self.fileInfo.file_path().split('.')[-1].lower()
        if extension in ('h5', 'hdf5', 'nx', 'nxs'):
            dictdump.dicttoh5(h5file=self.fileInfo.file_path(),
                              treedict=self.data,
                              h5path=self.fileInfo.data_path() or self.MATERIALS_DICT)
        elif extension in ('ini', 'cfg'):
            dictdump.dicttoini(ddict=self.data,
                               inifile=self.fileInfo.file_path())
        else:
            raise ValueError('Extension type (%s) not managed' % extension)


class Materials(object):
    """Define a matrix with int values. Each value is associated with a fisx
     material
    """
    def __init__(self, materials=None, matComposition=None):
        assert(materials is None or isinstance(materials, MaterialsDic))
        assert(matComposition is None or isinstance(matComposition, DataStored))

        self.materials = materials
        if materials is None:
            self.materials = MaterialsDic()
        self.matComposition = matComposition
        if self.matComposition is None:
            self.matComposition = MatComposition()

    def __eq__(self, other):
        return isinstance(other, Materials) and \
               self.materials == other.materials and \
               self.matComposition == other.matComposition

    def loadData(self):
        """

        """
        def compatibilityPy2Py3():
            # insure python 2 - python 3 compatibility
            mat = {}
            for key in self.materials.data:
                mat[key] = self.materials.data[key]
                if 'Comment' in mat[key] and hasattr(mat[key]['Comment'], 'astype'):
                    mat[key]['Comment'] = mat[key]['Comment'].astype('U13')
                if "CompoundList" in mat[key] and hasattr(mat[key]["CompoundList"], 'astype'):
                    mat[key]["CompoundList"] = mat[key]["CompoundList"].astype('U13')

            self.materials.data = mat

        if(self.materials is not None and
           FreeARTFileInfo._equalOrNone(self.materials.fileInfo, None) is False and
           self.materials.fileInfo.is_valid()):
            self.materials.loadData()
            if self.materials.data:
                compatibilityPy2Py3()

        if(self.matComposition is not None and
                   FreeARTFileInfo._equalOrNone(self.matComposition.fileInfo,
                                                None) is False and
               self.matComposition.fileInfo.is_valid()):
            self.matComposition.loadData()

    def save(self):
        """Save data using the information contained in the fileInfo"""
        if(self.materials is not None and
                   FreeARTFileInfo._equalOrNone(self.materials.fileInfo, None) is False):
            self.materials.save()
        if(self.matComposition is not None and
                   FreeARTFileInfo._equalOrNone(self.matComposition.fileInfo, None) is False):
            self.matComposition.save()

    def saveTo(self, file):
        """Save data using the information contained in the fileInfo or file
        if invalid"""
        self.materials.saveTo(file)
        self.matComposition.saveTo(file)

    def __str__(self):
        l = '  - materials is \n' + str(self.materials)
        l += '\n  - composition is \n' + str(self.matComposition)
        return l


class Detector(object):
    """
    Detector class

    :param float x:
    :param float y:
    :param float z:
    :param float width:
    """
    def __init__(self, x, y, z, width):
        self.x = x
        self.y = y
        self.z = z
        self.width = width

    def toDict(self):
        return {
            'det_width': self.width,
            'det_pos_x': self.x,
            'det_pos_y': self.y,
            'det_pos_z': self.z,
        }

    def getPosition(self):
        return (self.x, self.y, self.z)

    def __str__(self):
        res = 'x=%s\n' % self.x
        res += 'y=%s\n' % self.y
        res += 'z=%s\n' % self.z
        res += 'width=%s\n' % self.width
        return res

    def __eq__(self, other):
        if other is None:
            return False
        return self.x == other.x and self.y == other.y and self.z == other.z \
               and self.width == other.width



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/08/2016"


import os
import numpy
from numpy.distutils.misc_util import Configuration


def configuration(parent_package='', top_path=None):
    config = Configuration('core', parent_package, top_path)

    # =====================================
    # freeart core lib
    # =====================================
    config.add_extension(name='FreeART',
                         sources=['FreeART.pyx', os.path.join('src', 'GeometryFactory.cpp')],
                         include_dirs=['include', numpy.get_include()],
                         language='c++')

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(configuration=configuration)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Simple functions for fluorescence and transmission projections and reconstructions."""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/09/2016"

import freeart
from freeart.utils import genph, reconstrutils
import time
import numpy

############  Tx example #################


def produce_tx_SheppLogan_sinogram(_oversampling, _anglesNb, width, save=False):
    """Example of the generation of a simple 2D sinogram.
    parameters :
        - _oversampling : the number of oversampling number we want o take per mm covered by the ray
        - _anglesNb : the number of angle we want to make an acquisition for"""

    phGenerator = genph.PhantomGenerator()
    sheppLogan_phantom = phGenerator.get2DPhantomSheppLogan(n=width)
    if save:
        reconstrutils.savePhantom(sheppLogan_phantom, "ph_sheppLogan_256.edf")

    sheppLogan_phantom.shape = (sheppLogan_phantom.shape[0], sheppLogan_phantom.shape[1], 1)

    start_time = time.time()
    al = freeart.TxFwdProjection(sheppLogan_phantom, minAngle=0,
                                 maxAngle=2.0*numpy.pi, anglesNb=_anglesNb)
    al.setOverSampling(_oversampling)
    sinogram, angles = al.makeSinogram()

    if save:
        reconstrutils.saveSinogram(sinogram, "sino_sheppLogan_256.edf")

    execution_time = time.time() - start_time

    return sinogram, angles, execution_time

def make_tx_reconstruction(_sinogram, _angles, _nbIter, dampingFactor=0.02):
    """Example of the reconstruction from a sinogram and related data.
    parameters :
       - _sinogram : the sinogram to use for the reconstruction
       - _angles : the angles of acquisition registred on the sinogram
       - _nbIter : the number of iteration to execute for the reconstruction"""

    start_time = time.time()
    al = freeart.TxBckProjection(_sinogram, _angles)
    al.setDampingFactor(dampingFactor)

    res = al.iterate(_nbIter)
    execution_time = time.time() - start_time

    return res, execution_time

############  Fluo example #################


def produce_fluo_SheppLogan_sinogram(_oversampling, _anglesNb, _detSetup, _width, save=True):
    """Example of the generation of a simple 2D sinogram.
    parameters :
        - _oversampling : the number of oversampling number we want o take per mm covered by the ray
        - _anglesNb : the number of angle we want to make an acquisition for"""
    #generate the phantom
    phGenerator = genph.PhantomGenerator()
    sheppLogan_phantom = phGenerator.get2DPhantomSheppLogan(_width)
    sheppLogan_phantom.shape = (sheppLogan_phantom.shape[0], sheppLogan_phantom.shape[1], 1)

    start_time = time.time()

    #generate absMat and selfAnsMat
    absMat     = sheppLogan_phantom / 20.0 
    selfAbsMat = sheppLogan_phantom / 20.0 

    # rune the sinogram generation
    sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
        sheppLogan_phantom, absMat, selfAbsMat, _anglesNb, _detSetup, 10, True, False );

    if save :
        reconstrutils.savePhantom(sheppLogan_phantom, "ph_sheppLogan_64.edf")
        reconstrutils.savePhantom(absMat, "AM_sheppLogan_64.edf")
        reconstrutils.savePhantom(selfAbsMat, "SAM_sheppLogan_64.edf")

    execution_time = time.time() - start_time

    return sinogram, angles, absMat, selfAbsMat, execution_time


def make_fluo_reconstruction(_sinogram, _angles, _absMat, _selfAbsMat, _detSetup, _nbIter, _oversamp, _dampingFactor=0.05):
    """Example of the reconstruction from a sinogram and related data.
    parameters :
       - _sinogram : the sinogram to use for the reconstruction
       - _angles : the angles of acquisition registred on the sinogram
       - _nbIter : the number of iteration to execute for the reconstruction"""
    
    alRecons = freeart.FluoBckProjection(sinoDat=_sinogram,
                                         sinoAngles=_angles,
                                         expSetUp=_detSetup,
                                         absorp=_absMat,
                                         selfAbsorp=_selfAbsMat)

    alRecons.setOverSampling(_oversamp)
    alRecons.setDampingFactor(_dampingFactor)
    
    start_time = time.time()
    res = alRecons.iterate(_nbIter)
    execution_time = time.time() - start_time

    return res, execution_time




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""

"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

from . import configIO
import h5py
from . import config as configmod
import logging
import copy
try:
    from freeart import version
except:
    version = ''
import time
from silx.io import dictdump

logger = logging.getLogger(__name__)


class H5Struct():
    PROCESS_NAME = "freeart"
    RECONSTRUCTION_GRP_ID = 'reconstruction'
    CONFIGURATION_GRP_ID = 'configuration'

    def getOrCreateGroup(self, topLevelNode, group):
        def _getOrCreateNext(topLevelNode, nextGrp):
            if nextGrp in topLevelNode.keys():
                return topLevelNode[nextGrp]
            else:
                return topLevelNode.create_group(nextGrp)

        if group is None or group == "":
            return topLevelNode
        splits = group.lstrip('/').split('/')
        remainingPath, nextGrp = '/'.join(splits[1:]), splits[0]
        if len(splits) == 1:
            return _getOrCreateNext(topLevelNode, nextGrp)
        else:
            return self.getOrCreateGroup(
                topLevelNode=_getOrCreateNext(topLevelNode, nextGrp),
                group='/'.join(splits[1:]))

    def _h5FilePathIsValid(self, filePath):
        if filePath is None:
            return False
        splits = filePath.split("::")
        return len(splits) is 2 and len(splits[1]) > 0


class H5ConfigWriter(configIO.ConfigWriter, H5Struct):
    def write(self, filePath, reconsConfiguration, overwrite=False,
              merge=False):
        assert(isinstance(reconsConfiguration, configmod._ReconsConfig))
        confToSave = copy.copy(reconsConfiguration)
        if merge is True:
            confToSave.setFileInfoToH5File(filePath, save=True)

        file = h5py.File(filePath, mode='a')

        reconsGrp = self.getOrCreateGroup(file, self.RECONSTRUCTION_GRP_ID)
        reconsGrp.attrs['program'] = self.PROCESS_NAME
        reconsGrp.attrs['version'] = version
        reconsGrp.attrs['date'] = str(time.strftime("%d/%m/%Y"))

        # prepare configuration
        configGrp = self.getOrCreateGroup(reconsGrp, self.CONFIGURATION_GRP_ID)
        configGrp.attrs['format'] = 'ini'
        path = configGrp.name

        if overwrite is True:
            index = 0
            key = path + '/sino_file_' + str(index)
            while key in file:
                del file[key]
                index = index + 1
                key = path + '/sino_file_' + str(index)

        file.close()

        dictdump.dicttoh5(treedict=confToSave.toDict(),
                          h5file=filePath,
                          h5path=path,
                          mode='r+',
                          overwrite_data=overwrite)


class H5ConfigReader(configIO.ConfigReader, H5Struct):

    def read(self, filePath):
        dict = dictdump.load(filePath, fmat="hdf5")
        if self.RECONSTRUCTION_GRP_ID not in dict:
            raise IOError('structure of the hdf5 not recognize, can\'t '
                          'retrieve the configuration')
        if self.CONFIGURATION_GRP_ID not in dict[self.RECONSTRUCTION_GRP_ID]:
            raise IOError('structure of the hdf5 not recognize, can\'t '
                          'retrieve the configuration')

        configurationDict = dict[self.RECONSTRUCTION_GRP_ID][self.CONFIGURATION_GRP_ID]

        if 'general_settings' not in configurationDict:
            logger.warning('impossible to read reconstruction configuration,'
                           '`general_settings` not defined.'
                           'File structure is not valid')
            return None

        if 'reconstruction_type' not in configurationDict['general_settings']:
            logger.warning(
                'impossible to read reconstruction configuration,'
                '`general_settings/reconstruction_type` not defined.'
                'File structure is not valid')
            return None

        reconsType = configurationDict['general_settings']['reconstruction_type']
        if type(reconsType) is not str:
            reconsType = str(reconsType.tostring().decode())
        if reconsType not in self.dictReconsTypeToReconsConf:
            logger.error(
                'reconstruction type requested is not defined (%s)' % reconsType)
            return None
        conf = self.dictReconsTypeToReconsConf[reconsType]()._fromDict(configurationDict)

        self._reloadDatasets(config=conf)
        return conf
